#include "metainf.h"


/*
from   ecasound  

audioio-mp3.cpp
Language: C++
License: GPL
Copyright: (C) 1999-2005 Kai Vehmanen 
  
*/

/***************************************************************
 * Routines for parsing mp3 header information. Taken from XMMS
 * 1.2.5's mpg123 plugin.
 **************************************************************/

#define         MAXFRAMESIZE            1792

#define         MPG_MD_STEREO           0
#define         MPG_MD_JOINT_STEREO     1
#define         MPG_MD_DUAL_CHANNEL     2
#define         MPG_MD_MONO             3


int tabsel_123[2][3][16] =
  {
    {
      {0, 32, 64, 96, 128, 160, 192, 224, 256, 288, 320, 352, 384, 416, 448,},
      {0, 32, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320, 384,},
      {0, 32, 40, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320,}},

    {
      {0, 32, 48, 56, 64, 80, 96, 112, 128, 144, 160, 176, 192, 224, 256,},
      {0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160,},
      {0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160,}}
  };

long mpg123_freqs[9] =
  {44100, 48000, 32000, 22050, 24000, 16000, 11025, 12000, 8000};


static bool mpg123_head_check(unsigned long head)
{
  /* ref: http://www.mp3-tech.org/programmer/frame_header.html */

  /* frame sync must be 0xffe (11bits) */
  if ((head & 0xffe00000) != 0xffe00000) return false;
  /* layer must be non-null (2bits) */
  if (!((head >> 17) & 3)) return false;
  /* invalid bitrate index: all-ones (4bit) */
  if (((head >> 12) & 0xf) == 0xf) return false;
  /* invalid bitrate index: null (4bit) */
  if (!((head >> 12) & 0xf)) return false;
  /* invalid srate index: all-ones (2bit) */
  if (((head >> 10) & 0x3) == 0x3) return false;
#if 0
  /* invalid: mpeg2/2.5, layer I, protection bit off */
  if (((head >> 19) & 1) == 1 && ((head >> 17) & 3) == 3 && ((head >> 16) & 1) == 1) return false;
  /* - mpeg version 1, CRC protection bit */
  if ((head & 0xffff0000) == 0xfffe0000) return false;
#endif
	
  return true;
}



/*
 * the code a header and write the information
 * into the frame structure
 */
static bool mpg123_decode_header(struct frame *fr, unsigned long newhead)
{
  if (newhead & (1 << 20))
    {
      fr->lsf = (newhead & (1 << 19)) ? 0x0 : 0x1;
      fr->mpeg25 = 0;
    }
  else
    {
      fr->lsf = 1;
      fr->mpeg25 = 1;
    }
  fr->lay = 4 - ((newhead >> 17) & 3);
  if (fr->mpeg25)
    {
      fr->sampling_frequency = 6 + ((newhead >> 10) & 0x3);
    }
  else
    fr->sampling_frequency = ((newhead >> 10) & 0x3) + (fr->lsf * 3);
  fr->error_protection = ((newhead >> 16) & 0x1) ^ 0x1;

  if (fr->mpeg25)		/* allow Bitrate change for 2.5 ... */
    fr->bitrate_index = ((newhead >> 12) & 0xf);

  fr->bitrate_index = ((newhead >> 12) & 0xf);
  fr->padding = ((newhead >> 9) & 0x1);
  fr->extension = ((newhead >> 8) & 0x1);
  fr->mode = ((newhead >> 6) & 0x3);
  fr->mode_ext = ((newhead >> 4) & 0x3);
  fr->copyright = ((newhead >> 3) & 0x1);
  fr->original = ((newhead >> 2) & 0x1);
  fr->emphasis = newhead & 0x3;

  fr->stereo = (fr->mode == MPG_MD_MONO) ? 1 : 2;

  if (!fr->bitrate_index) {
    qDebug() << "Invalid bitrate!";
    return false;
  }

  int ssize = 0;
  switch (fr->lay)
    {
    case 1:
      //  	    fr->do_layer = mpg123_do_layer1;
      //  	    mpg123_init_layer2();
      fr->framesize = (long) tabsel_123[fr->lsf][0][fr->bitrate_index] * 12000;
      fr->framesize /= mpg123_freqs[fr->sampling_frequency];
      fr->framesize = ((fr->framesize + fr->padding) << 2) - 4;
      break;
    case 2:
      //  	    fr->do_layer = mpg123_do_layer2;
      //  	    mpg123_init_layer2();
      fr->framesize = (long) tabsel_123[fr->lsf][1][fr->bitrate_index] * 144000;
      fr->framesize /= mpg123_freqs[fr->sampling_frequency];
      fr->framesize += fr->padding - 4;
      break;
    case 3:
      //  	    fr->do_layer = mpg123_do_layer3;
      if (fr->lsf)
	ssize = (fr->stereo == 1) ? 9 : 17;
      else
	ssize = (fr->stereo == 1) ? 17 : 32;
      if (fr->error_protection)
	ssize += 2;
      fr->framesize = (long) tabsel_123[fr->lsf][2][fr->bitrate_index] * 144000;
      fr->framesize /= mpg123_freqs[fr->sampling_frequency] << (fr->lsf);
      fr->framesize = fr->framesize + fr->padding - 4;
      break;
    default:
      return false;
    }

  if(fr->framesize > MAXFRAMESIZE) {
    qDebug() <<  "Invalid framesize!";
    return false;
  }

  return true;
}


static bool mpg123_detect_by_content(const char* filename, struct frame* frp)
{
  FILE *file;
  quint8 tmp[4]; /* room for the 32bit head */
  quint32 head = 0;
  bool data_left = true;
  bool header_found = false;
  size_t offset = 0;

  if((file = std::fopen(filename, "rb")) == NULL) {
    qDebug() << "Unable to open file " << filename << ".";
    data_left = false;
  }
  /* search for headers in the first 262kB of data */
  while(data_left == true && offset < (1<<18)) {
    /* octet-by-octet search */
    if (std::fread(tmp, 1, 1, file) != 1) {
      qDebug() <<  "End of mp3 file, no valid header data found.";
      data_left = false;
      break;
    }

    head <<= 8;
    head |= tmp[0]; 
    offset += 1;

    if (offset > 3) {
      /* verify the header and if ok, fetch mp3 parameters and store
	 them to 'frp' */
      if (mpg123_head_check(head) && mpg123_decode_header(frp, head)) {
	if (header_found == true) {
	  /* two headers found, stop searching */
	  data_left = false;
	}
	else {
	  /* after the first header is found, skip to the next 
	     valid frame to verify that the first frame is not 
	     dummy frame (id3 or something similar) */
	  if (std::fseek(file, frp->framesize, SEEK_CUR) != 0) {
	    data_left = false;
	  }
	  header_found = true;
	}
	qDebug() <<  "Found mp3 header at offset " << (offset - 4);
      }
    }
  }

  return header_found;
}


CMetaInfo::CMetaInfo()
{
  processed = false; 
}


QString CMP3MetaInfo::stereo_mode()
{
  if (fr.mode == MPG_MD_MONO)
     return QString ("Mono");
  else 
  if (fr.mode == MPG_MD_STEREO)
     return QString ("Stereo");
  else 
  if (fr.mode == MPG_MD_JOINT_STEREO)
     return QString ("Joint stereo");
  else 
  if (fr.mode == MPG_MD_DUAL_CHANNEL)
     return QString ("Dual channel");

  return QString();
}


void CMP3MetaInfo::process (const QString &fname)
{
  processed = mpg123_detect_by_content (fname.toAscii(), &fr);

  if (processed)
     {
      bitrate = tabsel_123[fr.lsf][fr.lay - 1][fr.bitrate_index] * 1000;
      sample_rate = mpg123_freqs[fr.sampling_frequency];
     }
}


Cid3v1::Cid3v1()
{
  processed = false;
  
  hash.insert (0, "Blues");
  hash.insert (1, "Classic Rock");
  hash.insert (2, "Country");
  hash.insert (3, "Dance");
  hash.insert (4, "Disco");
  hash.insert (5, "Funk");
  hash.insert (6, "Grunge");
  hash.insert (7, "Hip-Hop");
  hash.insert (8, "Jazz");
  hash.insert (9, "Metal");
  hash.insert (10, "New Age");
  hash.insert (11, "Oldies");
  hash.insert (12, "Other");
  hash.insert (13, "Pop");
  hash.insert (14, "R&B");
  hash.insert (15, "Rap");
  hash.insert (16, "Reggae");
  hash.insert (17, "Rock");
  hash.insert (18, "Techno");
  hash.insert (19, "Industrial");
  hash.insert (20, "Alternative");
  hash.insert (21, "Ska");
  hash.insert (22, "Death Metal");
  hash.insert (23, "Pranks");
  hash.insert (24, "Soundtrack");
  hash.insert (25, "Euro-Techno");
  hash.insert (26, "Ambient");
  hash.insert (27, "Trip-Hop");
  hash.insert (28, "Vocal");
  hash.insert (29, "Jazz+Funk");
  hash.insert (30, "Fusion");
  hash.insert (31, "Trance");
  hash.insert (32, "Classical");
  hash.insert (33, "Instrumental");
  hash.insert (34, "Acid");
  hash.insert (35, "House");
  hash.insert (36, "Game");
  hash.insert (37,"Sound Clip");
  hash.insert (38, "Gospel");
  hash.insert (39, "Noise");
  hash.insert (40, "AlternRock");
  hash.insert (41, "Bass");
  hash.insert (42, "Soul");
  hash.insert (43, "Punk");
  hash.insert (44, "Space");
  hash.insert (45, "Meditative");
  hash.insert (46, "Instrumental Pop");
  hash.insert (47, "Instrumental Rock");
  hash.insert (48, "Ethnic");
  hash.insert (49, "Gothic");
  hash.insert (50, "Darkwave");
  hash.insert (51, "Techno-Industrial");
  hash.insert (52, "Electronic");
  hash.insert (53, "Pop-Folk");
  hash.insert (54, "Eurodance");
  hash.insert (55, "Dream");
  hash.insert (56, "Southern Rock");
  hash.insert (57, "Comedy");
  hash.insert (58, "Cult");
  hash.insert (59, "Gangsta");
  hash.insert (60, "Top 40");
  hash.insert (61, "Christian Rap");
  hash.insert (62, "Pop/Funk");
  hash.insert (63, "Jungle");
  hash.insert (64, "Native American");
  hash.insert (65, "Cabaret");
  hash.insert (66, "New Wave");
  hash.insert (67, "Psychadelic");
  hash.insert (68, "Rave");
  hash.insert (69, "Showtunes");
  hash.insert (70, "Trailer");
  hash.insert (71, "Lo-Fi");
  hash.insert (72, "Tribal");
  hash.insert (73, "Acid Punk");
  hash.insert (74, "Acid Jazz");
  hash.insert (75, "Polka");
  hash.insert (76, "Retro");
  hash.insert (77, "Musical");
  hash.insert (78, "Rock & Roll");
  hash.insert (79, "Hard Rock");
  hash.insert (80, "Folk");
  hash.insert (81, "Folk-Rock");
  hash.insert (82, "National Folk");
  hash.insert (83, "Swing");
  hash.insert (84, "Fast Fusion");
  hash.insert (85, "Bebob");
  hash.insert (86, "Latin");
  hash.insert (87, "Revival");
  hash.insert (88, "Celtic");
  hash.insert (89, "Bluegrass");
  hash.insert (90, "Avantgarde");
  hash.insert (91, "Gothic Rock");
  hash.insert (92, "Progressive Rock");
  hash.insert (93, "Psychedelic Rock");
  hash.insert (94, "Symphonic Rock");
  hash.insert (95, "Slow Rock");
  hash.insert (96, "Big Band");
  hash.insert (97, "Chorus");
  hash.insert (98, "Easy Listening");
  hash.insert (99, "Acoustic");
  hash.insert (100, "Humour");
  hash.insert (101, "Speech");
  hash.insert (102, "Chanson");
  hash.insert (103, "Opera");
  hash.insert (104, "Chamber Music");
  hash.insert (105, "Sonata");
  hash.insert (106, "Symphony");
  hash.insert (107, "Booty Bass");
  hash.insert (108, "Primus");
  hash.insert (109, "Porn Groove");
  hash.insert (110, "Satire");
  hash.insert (111, "Slow Jam");
  hash.insert (112, "Club");
  hash.insert (113, "Tango");
  hash.insert (114, "Samba");
  hash.insert (115, "Folklore");
  hash.insert (116, "Ballad");
  hash.insert (117, "Power Ballad");
  hash.insert (118, "Rhythmic Soul");
  hash.insert (119, "Freestyle");
  hash.insert (120, "Duet");
  hash.insert (121, "Punk Rock");
  hash.insert (122, "Drum Solo");
  hash.insert (123, "A capella");
  hash.insert (124, "Euro-House");
  hash.insert (125, "Dance Hall");
}


void Cid3v1::process (const QString &fname)
{
  processed = false;
 
  QFile file (fname);
  if (! file.open (QIODevice::ReadOnly))
     return;
     
  QDataStream in (&file); 
  in.skipRawData (file.size() - 128);
  
  in.readRawData (chr_signature, 3);
  
  if (chr_signature[0] == 'T' && 
      chr_signature[1] == 'A' && 
      chr_signature[2] == 'G'  
      )
      processed = true;
  else
      return;
             
  in.readRawData (chr_title, 30);
  in.readRawData (chr_artist, 30);
  in.readRawData (chr_album, 30);
  in.readRawData (chr_year, 4);
  in.readRawData (chr_comment, 30);
  in.readRawData (&chr_genre, 1);
    
  chr_title[30] = '\0';
  chr_artist[30] = '\0';
  chr_album[30] = '\0';  
  chr_year[4] = '\0';
  chr_comment[30] = '\0';
  
  fields.clear();

  QTextCodec::setCodecForCStrings (QTextCodec::codecForName (charset.toLatin1()));
  
  
  fields.insert ("ALBUM", chr_album);
  fields.insert ("ARTIST", chr_artist);
  fields.insert ("YEAR", chr_year);
  fields.insert ("COMMENT", chr_comment);
  fields.insert ("TITLE", chr_title);
  fields.insert ("GENRE", hash[chr_genre]);
 
  file.close();
}



#define ID3_GET_SIZE28(a, b, c, d)		\
(((a & 0x7f) << (24 - 3)) |			\
 ((b & 0x7f) << (16 - 2)) |			\
 ((c & 0x7f) << ( 8 - 1)) |			\
 ((d & 0x7f)))


void Cid3v2::process (const QString &fname)
{
  processed = false;
 
  QFile file (fname);
  if (! file.open (QIODevice::ReadOnly))
     return;
     
  QDataStream in (&file); 
 
  QTextCodec::setCodecForCStrings (QTextCodec::codecForName (charset.toLatin1()));
  
  
  char chr_signature[3];

  in.readRawData (chr_signature, 3);
  
  if (chr_signature[0] == 'I' && 
      chr_signature[1] == 'D' && 
      chr_signature[2] == '3'  
      )
      processed = true;
  else
      return;
  
    
  in >> ver_major;
  in >> ver_minor;
  
 // qDebug() << ver_major << "." << ver_minor; //hex
     
  char flags;
  in.readRawData (&flags, 1);
/*  
  if (flags & (1 << 7))
     qDebug() << "Unsynchronisation is used";
  
  if (flags & (1 << 6))
     qDebug() << "Extended header";
 
  if (flags & (1 << 5))
     qDebug() << "Experimental indicator";
*/
  char ID3v2_size[4];
  in.readRawData (ID3v2_size, 4);

  int iID3v2_size = ID3_GET_SIZE28 (ID3v2_size[0],
                                    ID3v2_size[1],
                                    ID3v2_size[2],
                                    ID3v2_size[3]);
  
  if (flags & (1 << 6)) //extended header is present
     {  
      char ext_header_size[4];
      in.readRawData (ext_header_size, 4);
     
      int iext_header_size = ID3_GET_SIZE28 (ext_header_size[0],
                                            ext_header_size[1], 
                                            ext_header_size[2], 
                                            ext_header_size[3]);
      //just skip this
      in.skipRawData (iext_header_size);
     }
  
  
	//Parse frames
	 
 int c = 0;
 while (c < iID3v2_size)
       {
        char frame_id[5];
        int x = in.readRawData (frame_id, 4);
        c += x; 
       
        frame_id[4] = '\0';
        
       // qDebug() << frame_id;
         
        quint32 frame_size;
        
        in >> frame_size;
        c += 4; 
       
//      qDebug() << "frame size = " << frame_size;
       
        quint16 frame_flags;
        in >> frame_flags;
        c += 2;
       
        if (frame_size < 1 || frame_size > 1000000)
            break;
        else
            if (frame_size > 1)
               {
                if (frame_id[0] == 'T')
                   {         
                    char enc;
                    x = in.readRawData (&enc, 1);
                    c += x; 
                    }
                else 
                   if (strcmp (frame_id, "COMM") == 0)
                      {
                       char enc;
                       char lang[3];
                  
                       x = in.readRawData (&enc, 1);
                       c += x; 
                    
                       x = in.readRawData (lang, 3);
                       c += x; 
                      
                       char t = 0;
                       do 
                         {
                          x = in.readRawData (&t, 1);
                          c++; 
                         }
                       
                       while (t != 0); 
                       frame_size -= 4; 
                       
                      }

                char* frame_data = new char[frame_size];
                x = in.readRawData (frame_data, frame_size - 1);
                c += x; 
                frame_data[frame_size - 1] = '\0';
  //              qDebug() << "frame_data" << frame_data;
                                
                fields.insert (frame_id, frame_data);
                
                delete (frame_data);
               }
           else  
               if (frame_size == 1)
                  {
                   in.skipRawData (1);
                   c++;
                  } 
          } 
  
  file.close();
  
  QString s;
  s = fields["TPE1"];
  if (! s.isEmpty()) 
     fields.insert ("ARTIST", s);

  s = fields["TIT2"];
  if (! s.isEmpty()) 
     fields.insert ("TITLE", s);
  
  s = fields["COMM"];
  if (! s.isEmpty()) 
     fields.insert ("COMMENT", s);

  s = fields["TYER"];
  if (! s.isEmpty()) 
     fields.insert ("YEAR", s);

  s = fields["TALB"];
  if (! s.isEmpty()) 
     fields.insert ("ALBUM", s);

  s = fields["TCON"];
  if (! s.isEmpty()) 
     fields.insert ("GENRE", s);
}

/*
void CFLAC::process (const QString &fname)
{
  QFile file (fname);
  if (! file.open (QIODevice::ReadOnly))
     return;
 
  QDataStream in (&file); 
 
  char chr_signature[4];

  in.readRawData (chr_signature, 4);
  
  if (chr_signature[0] == 'f' && 
      chr_signature[1] == 'L' && 
      chr_signature[2] == 'a' &&
      chr_signature[3] == 'C'  
      )
      processed = true;
  else
      return;
 
//metadata block hdr

  //char header[4];
//  in.readRawData (header, 4);
    

  // bits:
  // <1> Last-metadata-block flag
  // <7> BLOCK_TYPE
  //    0 : STREAMINFO
  //    1 : PADDING
  //    ..
  //    4 : VORBIS_COMMENT
  //    ..
  // <24> Length of metadata to follow

  char header[4];
  in.readRawData (header, 4);
  
  
  unsigned char blockType = header[0] & 0x7f;
  bool isLastBlock = (header[0] & 0x80) != 0;
  
    unsigned int val=0;
  
    val = val | header[1];
    val = val << 8;
    val = val | header[2];
    val = val << 8;
    val = val | header[3];

    qDebug() << val; 

    
  
  file.close();
}
*/

bool str_check (char *s1, char *s2, int size)
{
  for (int i = 0; i < size; i++)
       if (s1[i] != s2[i])
          return false;
   
  return true;
}


bool CWavReader::get_info (const QString fname)
{
  QFile fl (fname);
  
  if (! fl.open(QIODevice::ReadOnly))
     return false;
  
  QDataStream ds (&fl);

  t_wav_chunk_hdr hdr;

  while (! ds.atEnd())
        {
         ds.readRawData ((char *)&hdr, sizeof (hdr));
         
         if (str_check (hdr.chunk_id, "RIFF", 4))
            {
             char riff_type[4];
             ds.readRawData ((char *)&riff_type, sizeof (riff_type));
           } 
         else
         if (str_check (hdr.chunk_id, "fmt ", 4))
            {
             ds.readRawData ((char *)&wav_chunk_fmt, sizeof (wav_chunk_fmt));
             if (wav_chunk_fmt.bits_per_sample != 16)
                return false; 
            } 
         else
         if (str_check (hdr.chunk_id, "data", 4))
            {
             int nsamples = hdr.chunk_size / (wav_chunk_fmt.bits_per_sample / 8);
             
             double sqr_sum = 0.0;

             qint16 *ch_both = new qint16[nsamples];
             ds.readRawData ((char *)ch_both, hdr.chunk_size);
             for (int i = 0; i < nsamples; i++)
                  sqr_sum += (ch_both[i] * ch_both[i]);
         
             double srms = sqrt (sqr_sum / nsamples);
             rms = 20 * log10 (srms / SHRT_MAX);
                          
             delete [] ch_both;
             return true;
            }
        else
            ds.skipRawData (hdr.chunk_size); 
       }
  
  return false;  
}
